import logging

import numpy as np
import pandas as pd

from cycling_dynamics.calc import haversine_distance

# Constants
LAT_COL = "position_lat"
LON_COL = "position_long"
DIST_COL = "distance"
TIME_COL = "timestamp"


def create_grouped_segments(tracks: list[pd.DataFrame], start_distance: float, length: float) -> list[pd.DataFrame]:
    """Compare multiple activities over a segment.

    Args:
        tracks: List of 2 or more activity DataFrames. tracks[0] is treated as the control/base track.
        start_distance: The "distance" within track[0] to use as the control point.
        length: The length of the segment from the control points.

    Returns:
        List of trimmed and aligned DataFrames for each track.

    Raises:
        ValueError: If tracks list is empty or contains invalid DataFrames.

    """
    if not tracks or not all(isinstance(t, pd.DataFrame) for t in tracks):
        raise ValueError("tracks must be a non-empty list of pandas DataFrames")

    required_columns = [LAT_COL, LON_COL, DIST_COL, TIME_COL]
    for i, track in enumerate(tracks):
        if not all(col in track.columns for col in required_columns):
            raise ValueError(f"Track {i} is missing required columns. Required: {required_columns}")

    control = tracks[0]
    start_idx, end_idx = get_segment_indices(control, start_distance, length)
    start_point = control.iloc[start_idx][[LAT_COL, LON_COL, DIST_COL]].values
    end_point = control.iloc[end_idx][[LAT_COL, LON_COL, DIST_COL]].values

    logging.info(f"Start point lat, lon, distance: {start_point}")
    logging.info(f"End point lat, lon, distance: {end_point}")

    trimmed_tracks = [process_control_track(control, start_idx, end_idx, start_point[2])]

    for i, track in enumerate(tracks[1:], 1):
        trimmed_track = process_track(track, start_point, end_point, i)
        trimmed_tracks.append(trimmed_track)

    return trimmed_tracks


def get_segment_indices(track: pd.DataFrame, start_distance: float, length: float) -> tuple[int, int]:
    start_idx = (track[DIST_COL] - start_distance).abs().idxmin()
    end_idx = (track[DIST_COL] - (start_distance + length)).abs().idxmin()
    return start_idx, end_idx


def process_control_track(track: pd.DataFrame, start_idx: int, end_idx: int, start_distance: float) -> pd.DataFrame:
    control = track.loc[start_idx:end_idx].copy().reset_index(drop=True)
    control[DIST_COL] -= start_distance
    control["seconds"] = (control[TIME_COL] - control[TIME_COL].min()).dt.total_seconds()
    control["ride"] = 0
    return control


def process_track(track: pd.DataFrame, start_point: np.ndarray, end_point: np.ndarray, ride_num: int) -> pd.DataFrame:
    logging.info(f"Processing Track {ride_num}")

    start_idx, end_idx = find_matching_points(track, start_point, end_point)

    trimmed_track = track.loc[start_idx:end_idx].copy().reset_index(drop=True)
    trimmed_track[DIST_COL] -= trimmed_track[DIST_COL].iloc[0]
    trimmed_track["ride"] = ride_num
    trimmed_track["seconds"] = (trimmed_track[TIME_COL] - trimmed_track[TIME_COL].min()).dt.total_seconds()

    return trimmed_track


def find_matching_points(track: pd.DataFrame, start_point: np.ndarray, end_point: np.ndarray) -> tuple[int, int]:
    start_idx = track.apply(
        lambda row: haversine_distance(start_point[0], start_point[1], row[LAT_COL], row[LON_COL]), axis=1
    ).idxmin()
    end_idx = track.apply(
        lambda row: haversine_distance(end_point[0], end_point[1], row[LAT_COL], row[LON_COL]), axis=1
    ).idxmin()
    return start_idx, end_idx


# Assuming haversine_distance function is defined elsewhere



import pytest

from cycling_dynamics.load_data import load_fit_file
from cycling_dynamics.segments import match_segments

FIT_FILES = [
    "alex_24HOP_2024-02-17-185919-ELEMNT ROAM 3FF5-53-0.fit",
    "alex_24HOP_2024-02-17-185919-ELEMNT_ROAM_3FF5-53-0.fit",
    "alex_24HOP_2024-02-18-002255-ELEMNT ROAM 3FF5-54-0.fit",
    "alex_24HOP_2024-02-18-002255-ELEMNT_ROAM_3FF5-54-0.fit",
    "vincent_lap_1_24HOP_14012433014_ACTIVITY.fit",
    "vincent_lap_2_24HOP_14010180820_ACTIVITY.fit",
]
TEST_DATA_DIR = "test_data"


@pytest.fixture()
def get_tracks():
    """Load all the tacks for the test."""
    return [load_fit_file(f"{TEST_DATA_DIR}/{f}") for f in FIT_FILES]


def test_match_segments(get_tracks):
    """Basic tests for matching segments across tracks."""
    matched = match_segments(tracks=get_tracks, start_distance=5000, length=10000)
    # print(f"\n{matched[0][["seconds", "distance", "speed", "position_lat"]]}")
    # print(min([m.distance.max() for m in matched]), max([m.distance.max() for m in matched]))
    assert abs(min([m.distance.max() for m in matched]) - max([m.distance.max() for m in matched])) <= 15



"""Python module to contain base calculating functions."""

import logging
from math import asin, atan, cos, radians, sin, sqrt, tan

import numpy as np
import pandas as pd


def haversine_distance(lat1: float, lon1: float, lat2: float, lon2: float) -> float:
    """Calculate great circle distance between two points in a sphere, given longitudes and latitudes.

     Referance: https://en.wikipedia.org/wiki/Haversine_formula.

    We know that the globe is "sort of" spherical, so a path between two points
    isn't exactly a straight line. We need to account for the Earth's curvature
    when calculating distance from point A to B. This effect is negligible for
    small distances but adds up as distance increases. The Haversine method treats
    the earth as a sphere which allows us to "project" the two points A and B
    onto the surface of that sphere and approximate the spherical distance between
    them. Since the Earth is not a perfect sphere, other methods which model the
    Earth's ellipsoidal nature are more accurate but a quick and modifiable
    computation like Haversine can be handy for shorter range distances.

    Args:
        lat1, lon1: latitude and longitude of coordinate 1
        lat2, lon2: latitude and longitude of coordinate 2
    Returns:
        geographical distance between two points in metres
    >>> from collections import namedtuple
    >>> point_2d = namedtuple("point_2d", "lat lon")
    >>> SAN_FRANCISCO = point_2d(37.774856, -122.424227)
    >>> YOSEMITE = point_2d(37.864742, -119.537521)
    >>> f"{haversine_distance(*SAN_FRANCISCO, *YOSEMITE):0,.0f} meters"
    '254,352 meters'

    """
    # CONSTANTS per WGS84 https://en.wikipedia.org/wiki/World_Geodetic_System
    # Distance in metres(m)
    AXIS_A = 6378137.0
    AXIS_B = 6356752.314245
    RADIUS = 6378137
    # Equation parameters
    # Equation https://en.wikipedia.org/wiki/Haversine_formula#Formulation
    flattening = (AXIS_A - AXIS_B) / AXIS_A
    phi_1 = atan((1 - flattening) * tan(radians(lat1)))
    phi_2 = atan((1 - flattening) * tan(radians(lat2)))
    lambda_1 = radians(lon1)
    lambda_2 = radians(lon2)
    # Equation
    sin_sq_phi = sin((phi_2 - phi_1) / 2)
    sin_sq_lambda = sin((lambda_2 - lambda_1) / 2)
    # Square both values
    sin_sq_phi *= sin_sq_phi
    sin_sq_lambda *= sin_sq_lambda
    h_value = sqrt(sin_sq_phi + (cos(phi_1) * cos(phi_2) * sin_sq_lambda))
    return 2 * RADIUS * asin(h_value)


def angle_type(a, b, c):
    """Calculate the cosine of the angle using Law of Cosines and determine the angle type.

    If the angle between sides a and b acute, obtuse or 90 degrees (right angle)
     a: The original control point to next point on path
     b: The original control point to point on other path
     c: Second point on path the other path point.
    """
    angle = (a**2 + b**2 - c**2) / (2 * a * b)

    # Determine the angle type
    if angle > 0:  # After control point
        return "after"
    elif angle < 0:  # Before control point
        return "before"
    else:
        return "beside"  # cos_C == 0 implies a right angle


def add_metrics(df: pd.DataFrame, rolling_window: int = 30, ftp: int | None = None) -> pd.DataFrame:
    """Add metrics to the dataframe"""
    logging.info("Adding metrics")
    # Speed
    df[f"speed {rolling_window}sec"] = df["speed"].rolling(window=rolling_window).mean()

    # Efficiency
    df[f"speed per watt {rolling_window}sec"] = (
        df["speed"].rolling(window=rolling_window).mean() / df["power"].rolling(window=rolling_window).mean()
    )
    df[f"speed sqrd per watt {rolling_window}sec"] = (
        df["speed"].rolling(window=rolling_window).mean() ** 2 / df["power"].rolling(window=rolling_window).mean()
    )

    # power
    df["np"] = (df["power"] ** 4).rolling(window=30).mean() ** 0.25
    if ftp is not None:
        df = intensity_factor(df, ftp=ftp)
        df = total_training_stress(df, ftp=ftp)
        logging.info(f"Total Training Stress (TSS): {df['TSS'].iloc[-1]:0.2f} kcal/hr (FTP: {ftp})")
    return df


def total_training_stress(df: pd.DataFrame, ftp: int) -> pd.DataFrame:
    """Calculat Total Training Stress."""
    df["TSS"] = (df["power"] * df["IF"] * df["seconds"] / ftp / 3600).cumsum()
    return df


def normalized_power(df: pd.DataFrame) -> pd.DataFrame:
    """Calculate the normalized power of a ride."""
    df["np"] = (df["power"] ** 4).rolling(window=30).mean() ** 0.25
    return df


def intensity_factor(df: pd.DataFrame, ftp: int) -> pd.DataFrame:
    """Calculate the intensity factor of a ride."""
    df["IF"] = ((df["power"] ** 4).rolling(window=30).mean() ** 0.25) / ftp
    return df


def vam(df: pd.DataFrame) -> pd.DataFrame:
    """Calculate vertical ascent rate."""
    # TODO Improve calculation by using 3 points
    df["vam"] = (df["altitude"].diff() / df.seconds.diff()) * 3600
    return df


def slope(df: pd.DataFrame) -> pd.DataFrame:
    """Calculate the slope."""
    # TODO Improve calculation by using 3 points
    df["slope"] = df["altitude"].diff() / df.distance.diff()
    df["slope_3sec"] = df["slope"].rolling(window=3, center=True).mean()
    df["slope_3sec"] = df["slope_3sec"].fillna(df["slope_3sec"])
    return df


def zero_seconds(df: pd.DataFrame) -> pd.DataFrame:
    """Create or reset seconds callumn starting at zero."""
    # TODO: make sure it is sorted
    df["seconds"] = df["timestamp"].sub(df["timestamp"].min()).dt.total_seconds()
    # if "seconds" not in df.columns:
    #     df["seconds"] = pd.to_datetime(df.index, unit="s", origin="unix").astype(int) // 10**9
    # df["seconds"] = df["seconds"] - df.seconds.min()
    return df


def speed(df: pd.DataFrame) -> pd.DataFrame:
    """Calculate speed if missing."""
    # TODO Improve calculation by using 3 points
    df["speed"] = df.distance.diff() / df.time.diff()

def speed_3sec(df: pd.DataFrame) -> pd.DataFrame:
    df["speed_3sec"] = df["speed"].rolling(window=3, center=True).mean()
    df['speed_3sec'] = df['speed_moving_avg'].fillna(df['speed'])
    return df


def air_density(df: pd.DataFrame, temperature: float = 30) -> pd.DataFrame:
    """Calculate air density."""
    if temperature not in df.columns:
        logging.info("Using default temperature of 30 degrees C")
        df["temperature"] = 30
    df["air_density"] = (
        (101325 / (287.05 * 273.15))
        * (273.15 / (df["temperature"] + 273.15))
        * np.exp((-101325 / (287.05 * 273.15)) * 9.8067 * (df["altitude"] / 101325))
    )
    return df



import logging

import pandas as pd
from garmin_fit_sdk import Decoder, Stream

from cycling_dynamics import calc
from cycling_dynamics.calc import slope

logging.basicConfig(level=logging.INFO)


def load_fit_file(
    file_path: str, add_fields: tuple[str] = ("seconds", "slope", "vam", "speed", "normalized_power", "air_density")
) -> pd.DataFrame:
    """Load a FIT file and return its data as a pandas DataFrame.

    Args:
        file_path (str): Path to the FIT file.
        add_fields (bool, optional): Flag to add additional metrics. Defaults to True.

    Returns:
        pd.DataFrame: DataFrame containing the FIT file data.

    Raises:
        ValueError: If the file is not a valid FIT file or if there are decoding errors.

    """
    logging.info("Loading FIT file")

    stream = Stream.from_file(file_path)
    decoder = Decoder(stream)

    if not decoder.is_fit():
        logging.error(f"File is not FIT: {file_path}")
        raise ValueError(f"File is not a valid FIT file: {file_path}")

    messages, errors = decoder.read()

    if errors:
        logging.error(f"Errors: {errors}")
        raise ValueError(f"Errors occurred while decoding the FIT file: {errors}")

    df = pd.DataFrame(messages["record_mesgs"])

    if 'altitude' not in df.columns:
        df['altitude'] = 0

    # Convert lat/long to decimal degrees
    for col in ["position_lat", "position_long"]:
        if col in df.columns:
            df[col] = df[col] / 1e7

    enhanced_cols = ["enhanced_distance", "enhanced_speed", "enhanced_altitude"]
    for col in enhanced_cols:
        if col in df.columns:
            logging.info(f"Using {col}")
            if col.split("_")[1] in df.columns:
                df.drop(columns=[col.split("_")[1]], inplace=True)
            df.rename(columns={col: col.split("_")[1]}, inplace=True)

    if "seconds" in add_fields and "seconds" not in df.columns:
        logging.info("Using calculated seconds")
        df = calc.zero_seconds(df)
    if "speed" in add_fields and "speed" not in df.columns:
        logging.info("Using calculated speed")
        df = calc.speed(df)
    if "slope" in add_fields and "slope" not in df.columns:
        logging.info("Using calculated slope")
        df = slope(df)
    if "vam" in add_fields and "vam" not in df.columns:
        logging.info("Using calculated vertical acceleration magnitude")
        df = calc.vam(df)
    if "normalized_power" in add_fields and "normalized_power" not in df.columns:
        logging.info("Using calculated normalized power")
        df = calc.normalized_power(df)
    if "air_density" in add_fields and "air_density" not in df.columns:
        logging.info("Using calculated air density")
        df = calc.air_density(df)

    return df



"""Dynamic simulations."""

import numpy as np
import pandas as pd


def simulator(
    df: pd.DataFrame,
    smoothing: int = 3,
    rider_weight: float = 65.0,
    bike_weight: float = 5.0,
    wind_speed: float = 0,
    wind_direction: int = 0,
    temperature: float = 30,
    drag_coefficient: float = 0.8,
    frontal_area: float = 0.565,
    rolling_resistance: float = 0.005,
    efficiency_loss: float = 0.04,
) -> pd.DataFrame:
    """Calculate the components of power loss at each point and estimated total power needed.

    df: Usually from a FIT file or other GPS file. Required columns [distance, speed and/or time, altitude]
    """
    try:
        assert all([c in df.columns for c in ["seconds", "distance", "altitude"]])
    except AssertionError:
        raise AssertionError("Missing columns in dataframe. Must have 'seconds', 'distance', and 'altitude'")

    CdA = drag_coefficient * frontal_area

    df["effective_wind_speed"] = np.cos(np.radians(wind_direction)) * wind_speed

    # Components of power, watts
    df["air_drag_watts"] = (
        0.5 * CdA * df["air_density"] * np.square(df["speed"] + df["effective_wind_speed"]) * df["speed"]
    )
    df["climbing_watts"] = (bike_weight + rider_weight) * 9.8067 * np.sin(np.arctan(df["slope"])) * df["speed"]
    df["rolling_watts"] = (
        np.cos(np.arctan(df["slope"])) * 9.8067 * (bike_weight + rider_weight) * rolling_resistance * df["speed"]
    )
    df["acceleration_watts"] = (bike_weight + rider_weight) * (df["speed"].diff() / df["seconds"].diff()) * df["speed"]
    df["est_power_no_loss"] = df[["air_drag_watts", "climbing_watts", "rolling_watts", "acceleration_watts"]].sum(
        axis="columns"
    )
    df["est_power"] = df["est_power_no_loss"] / (1 - efficiency_loss)
    df["efficiency_loss_watts"] = df["est_power_no_loss"] - df["est_power"]
    df["est_power_no_acceleration"] = (df["est_power_no_loss"] - df["acceleration_watts"]) / (1 - efficiency_loss)

    if smoothing > 0:
        df["speed_smoothed"] = df["speed"].rolling(window=smoothing, center=True).mean()
        df["slope_smoothed"] = df["slope"].rolling(window=smoothing, center=True).mean()
        df["power_smoothed"] = df["power"].rolling(window=smoothing, center=True).mean()
        df["air_drag_watts_smoothed"] = df["air_drag_watts"].rolling(window=smoothing, center=True).mean()
        df["climbing_watts_smoothed"] = df["climbing_watts"].rolling(window=smoothing, center=True).mean()
        df["rolling_watts_smoothed"] = df["rolling_watts"].rolling(window=smoothing, center=True).mean()
        df["est_power_smoothed"] = df["est_power"].rolling(window=smoothing, center=True).mean()
        df["efficiency_loss_watts_smoothed"] = df["efficiency_loss_watts"].rolling(window=smoothing, center=True).mean()
        df["acceleration_watts_smoothed"] = df["acceleration_watts"].rolling(window=smoothing, center=True).mean()
        df["est_power_no_acceleration_smoothed"] = (
            df["est_power_no_acceleration"].rolling(window=smoothing, center=True).mean()
        )

    return df



import os

import pandas as pd
import pytest

from src.cycling_dynamics.load_data import load_fit_file

# Assuming the FIT files are in a 'test_data' directory
FIT_FILES = [
    "alex_24HOP_2024-02-17-185919-ELEMNT ROAM 3FF5-53-0.fit",
    "alex_24HOP_2024-02-17-185919-ELEMNT_ROAM_3FF5-53-0.fit",
    "alex_24HOP_2024-02-18-002255-ELEMNT ROAM 3FF5-54-0.fit",
    "alex_24HOP_2024-02-18-002255-ELEMNT_ROAM_3FF5-54-0.fit",
    "vincent_lap_1_24HOP_14012433014_ACTIVITY.fit",
    "vincent_lap_2_24HOP_14010180820_ACTIVITY.fit",
    "vincent_mtb_id_surface_14517623900_ACTIVITY.fit",
]
TEST_DATA_DIR = "test_data"


@pytest.fixture(params=FIT_FILES)
def fit_file_path(request):
    return os.path.join(TEST_DATA_DIR, f"{request.param}")


def test_load_fit_file_success(fit_file_path):
    df = load_fit_file(fit_file_path)

    assert isinstance(df, pd.DataFrame)
    assert len(df) > 0

    # Check for expected columns
    expected_columns = ["timestamp", "position_lat", "position_long", "altitude", "speed"]
    for col in expected_columns:
        assert col in df.columns, f"Expected column {col} not found in DataFrame"

    # Check data types
    assert pd.api.types.is_datetime64_any_dtype(df["timestamp"])
    assert pd.api.types.is_float_dtype(df["position_lat"])
    assert pd.api.types.is_float_dtype(df["position_long"])
    assert pd.api.types.is_float_dtype(df["altitude"])
    assert pd.api.types.is_float_dtype(df["speed"])

    # Check value ranges
    assert df["position_lat"].min() >= -90
    assert df["position_lat"].max() <= 90
    assert df["position_long"].min() >= -180
    assert df["position_long"].max() <= 180
    assert df["speed"].min() >= 0


# def test_load_fit_file_add_metrics(fit_file_path):
#     df_with_metrics = load_fit_file(fit_file_path, add_metrics=True)
#     df_without_metrics = load_fit_file(fit_file_path, add_metrics=False)
#
#     assert len(df_with_metrics.columns) > len(df_without_metrics.columns)
#
#     # Check for additional metrics (adjust based on what metrics you add)
#     additional_metrics = ['speed_rolling_avg']  # Add more as you implement them
#     for metric in additional_metrics:
#         assert metric in df_with_metrics.columns
#         assert metric not in df_without_metrics.columns


# @pytest.mark.parametrize("rolling_window", [10, 30, 60])
# def test_load_fit_file_rolling_window(fit_file_path, rolling_window):
#     df = load_fit_file(fit_file_path, add_metrics=True, rolling_window=rolling_window)
#
#     # Check if rolling average is calculated correctly
#     if 'speed_rolling_avg' in df.columns:
#         # The first rolling_window - 1 values should be NaN
#         assert df['speed_rolling_avg'].iloc[:rolling_window-1].isna().all()
#         # The rest should not be NaN
#         assert not df['speed_rolling_avg'].iloc[rolling_window:].isna().any()


def test_load_fit_file_data_integrity(fit_file_path):
    df = load_fit_file(fit_file_path)

    # Check for monotonically increasing timestamps
    assert df["timestamp"].is_monotonic_increasing

    # Check for no duplicate timestamps
    assert not df["timestamp"].duplicated().any()

    # Check for reasonable speed values (e.g., below 100 m/s which is 360 km/h)
    assert df["speed"].max() < 100


if __name__ == "__main__":
    pytest.main([__file__])
